#-----------------------------
# -- Angola --
#-----------------------------

import copy
from . import lib 
from math import ceil
from typing import Any, List


# Map filter
# Map filter operators with their functions
QUERY_FILTERING_MAPPER = {
    '=': '_is_equal',
    'eq': '_is_equal',
    '!=': '_is_not_equal',
    'neq': '_is_not_equal',
    '>': '_is_greater',
    'gt': '_is_greater',
    '<': '_is_smaller',
    'lt': '_is_smaller',
    '>=': '_is_greater_equal',
    'gte': '_is_greater_equal',
    '<=': '_is_smaller_equal',
    'lte': '_is_smaller_equal',
    'in': '_is_in',
    'notin': '_is_not_in',
    'null': '_is_null',
    'notnull': '_is_not_null',
    'startswith': '_is_starts_with',
    'endswith': '_is_ends_with',
    'includes': '_is_includes',
    'between': "_is_between",
    'contains': '_is_contains',
    'exists': '_is_exists',
    'notexists': '_is_not_exists',

}

# FILTER_OPERATOR:
FILTER_OPERATOR = ":$"

# SQL_OPERATORS: valid/strict operators for SQL
SQL_OPERATORS = {
    "eq": "= ?",
    "ne": "!= ?",
    "lt": "< ?",
    "gt": "> ? ",
    "lte": "<= ?",
    "gte": ">= ? ",
    "between": "BETWEEN ? AND ?",
}

class _Undefined: pass

"""
DictQuery

This library helps filter sub document in a list of dict 

:USAGE

data = List[dict]
data = [
    {
        "location": "USA",
        "age": 21,
        "email": "something@this.io",
        "friends": [
            {
                "name": "Jacob",
                "city": "Charlotte",

            }
        ]
    },
    ...
]


filters = {
    "location": "USA",
    "age:$gte": 18,
    "email:$endswith": ".io",
    "friends[*].city:$in": ["Charlotte", "Atlanta"]
}

filtered_data = DictQuery(data).execute(filter)

"""
class DictQuery(object):

    def __init__(self, data: list):

        if not isinstance(data, list):
            raise TypeError("Provided Data is not list")
        self._data = self._flatten_data_list(data)

    def execute(self, filters) -> list:
        self._reset_queries()
        self._build_queries(filters)
        return self._unflatten_data_list(self._execute_queries(self._data, self._queries))
 
    def _flatten_data_list(self, data):
        return [lib.flatten_dict(d) for d in data] if isinstance(data, list) else []

    def _unflatten_data_list(self, data):
        return [lib.unflatten_dict(d) for d in data] if isinstance(data, list) else []

    def _reset_queries(self):
        """ 
        Reset previous query data 
        """
        self._queries = []
        self._current_query_index = 0

    def _store_query(self, query_items):
        """Make where clause

        :@param query_items
        :@type query_items: dict
        """
        temp_index = self._current_query_index
        if len(self._queries) - 1 < temp_index:
            self._queries.append([])
        self._queries[temp_index].append(query_items)
 
    def _execute_queries(self, data:list, queries:list):
        """
        Args:
            data:list
        """
        _matcher = _Matcher()
       
        def func(item):
            or_check = False
            for queriesx in queries:
                and_check = True
                for query in queriesx:
                    _qk = query.get("key")

                    # Filter sub-list using `[*].` 
                    # /basepath/something[*].x
                    if "[*]." in _qk:    
                        and_check2 = False                    
                        basepath, subpath = _qk.split("[*].")
                        subdata = item.get(basepath)
                        if subdata:
                            for subitem in subdata:
                                and_check2 |= _matcher._match(
                                    subitem.get(subpath, _Undefined),
                                    query.get('operator'),
                                    query.get('value'),
                                    query.get('case_insensitive')
                                )    
                        and_check &= and_check2
                    
                    # regular query
                    else:
                        and_check &= _matcher._match(
                            item.get(_qk, _Undefined),
                            query.get('operator'),
                            query.get('value'),
                            query.get('case_insensitive')
                        )
                    
                or_check |= and_check
            return or_check

        return list(filter(lambda item: func(item), data))

    def _where(self, key, operator, value, case_insensitive=False):
        """Make where clause

        :@param key
        :@param operator
        :@param value
        :@type key,operator,value: string

        :@param case_insensitive
        :@type case_insensitive: bool

        :@return self
        """
        self._store_query({
            "key": key,
            "operator": operator,
            "value": value,
            "case_insensitive": case_insensitive
        })

        return self

    def _or_where(self, key, operator, value):
        """Make _or_where clause

        :@param key
        :@param operator
        :@param value
        :@type key, operator, value: string

        :@return self
        """
        if len(self._queries) > 0:
            self._current_query_index += 1
        self._store_query({"key": key, "operator": operator, "value": value})
        return self

    def _build_queries(self, filters: dict) -> tuple:
        """
        Create a FILTER clause

        Params:
            filters: dict
                {
                    'name': 'something',
                    'age:$gt': 18,
                    'cities:$in': ['charlotte', 'Concord'],
                    '$or': [
                        {
                            "cities:$in": [],
                            "_perms.read:$in":[] 
                        }
                    ]
                }
        """

        for k in filters:
            if k.startswith("$"):
                k_ = k.lower()
                # operation
                if k_ in ["$or"] and isinstance(filters[k], (dict, list)):
                    fk = filters[k]
                    if isinstance(fk, dict):
                        fk = [fk]
                    for k0 in fk:
                        for k2 in k0:
                            self._build_query_row(k2, k0[k2], _or=True)
                else:
                    raise Exception("Invalid logic: %s" % k)
            else:
                self._build_query_row(k, filters[k])

    def _build_query_row(self, k: str, value: Any, _or: bool = False):
        operator = "$eq"  # default operator
        if ":" in k:
            k, operator = k.split(":", 2)
            operator = operator.lower()
        operator = operator.replace("$", "")

        if _or:
            self._or_where(k, operator, value)
        self._where(k, operator, value)

class _Matcher(object):
    """docstring for Helper."""
    MAP = QUERY_FILTERING_MAPPER

    def _is_equal(self, x, y):
        return x == y

    def _is_not_equal(self, x, y):
        return x != y

    def _is_greater(self, x, y):
        return x > y

    def _is_smaller(self, x, y):
        return x < y

    def _is_greater_equal(self, x, y):
        return x >= y

    def _is_smaller_equal(self, x, y):
        return x <= y

    def _is_in(self, key, arr):
        return isinstance(arr, list) and \
            bool(len(([k for k in key if k in arr]
                 if isinstance(key, list) else key in arr)))

    def _is_not_in(self, key, arr):
        return isinstance(arr, list) and (key not in arr)

    def _is_null(self, x, y=None):
        return x is None

    def _is_not_null(self, x, y=None):
        return x is not None

    def _is_starts_with(self, data, val):
        if not isinstance(data, str):
            return False
        return data.startswith(val)

    def _is_ends_with(self, data, val):
        if not isinstance(data, str):
            return False
        return data.endswith(val)

    def _is_includes(self, ldata, val):
        if isinstance(ldata, (list, dict, str)):
            return val in ldata
        return False

    def _is_between(self, data, val):
        # TODO: IMPLEMENT
        return False
        
    def _is_contains(self, data, val):
        # TODO: IMPLEMENT
        return False

    def _is_exists(self, data, val=None):
        return data is not _Undefined

    def _is_not_exists(self, data, val=None):
        return data is _Undefined



    def _to_lower(self, x, y):
        return [[v.lower() if isinstance(v, str) else v for v in val]
                if isinstance(val, list) else val.lower()
                if isinstance(val, str) else val
                for val in [x, y]]

    def _match(self, x, op, y, case_insensitive):
        """Compare the given `x` and `y` based on `op`

        :@param x, y, op, case_insensitive
        :@type x, y: mixed
        :@type op: string
        :@type case_insensitive: bool

        :@return bool
        :@throws ValueError
        """
        if (op not in self.MAP):
            raise ValueError('Invalid where condition given: %s' % op)

        if case_insensitive:
            x, y = self._to_lower(x, y)

        func = getattr(self, self.MAP.get(op))
        return func(x, y)


class Cursor(object):

    def __init__(self, cursordat: list, sort=None, offset=None, limit=None):
        """Initialize the mongo iterable cursor with data"""
        self.cursordat = cursordat or []
        self.cursorpos = -1

        if len(self.cursordat) == 0:
            self.currentrec = None
        else:
            self.currentrec = self.cursordat[self.cursorpos]

        if sort:
            self.sort(sort)

        self.paginate(offset=offset, limit=limit)

    def __getitem__(self, key):
        """Gets record by index or value by key"""
        if isinstance(key, int):
            return self.cursordat[key]
        return self.currentrec[key]

    def __len__(self):
        return self.count
        
    def paginate(self, offset=None, limit=None):
        """Paginate list of records"""
        if not self.count or not limit:
            return
        offset = offset or 0
        pages = int(ceil(self.count / float(limit)))
        limits = {}
        last = 0
        for i in range(pages):
            current = limit * i
            limits[last] = current
            last = current
        # example with count == 62
        # {0: 20, 20: 40, 40: 60, 60: 62}
        if limit and limit < self.count:
            limit = limits.get(offset, self.count)
            self.cursordat = self.cursordat[offset: limit]

        return self

    def limit(self, limit=None, offset=None):
        self.paginate(limit=limit, offset=offset)
        return self

    def _order(self, value, is_reverse=None):
        """Parsing data to a sortable form
        By giving each data type an ID(int), and assemble with the value
        into a sortable tuple.
        """

        def _dict_parser(dict_doc):
            """ dict ordered by:
            valueType_N -> key_N -> value_N
            """
            result = list()
            for key in dict_doc:
                data = self._order(dict_doc[key])
                res = (data[0], key, data[1])
                result.append(res)
            return tuple(result)

        def _list_parser(list_doc):
            """list will iter members to compare
            """
            result = list()
            for member in list_doc:
                result.append(self._order(member))
            return result

        # (TODO) include more data type
        if value is None or not isinstance(value, (dict,
                                                   list,
                                                   str,
                                                   bool,
                                                   float,
                                                   int)):
            # not support/sortable value type
            value = (0, None)

        elif isinstance(value, bool):
            value = (5, value)

        elif isinstance(value, (int, float)):
            value = (1, value)

        elif isinstance(value, str):
            value = (2, value)

        elif isinstance(value, dict):
            value = (3, _dict_parser(value))

        elif isinstance(value, list):
            if len(value) == 0:
                # [] less then None
                value = [(-1, [])]
            else:
                value = _list_parser(value)

            if is_reverse is not None:
                # list will firstly compare with other doc by it's smallest
                # or largest member
                value = max(value) if is_reverse else min(value)
            else:
                # if the smallest or largest member is a list
                # then compaer with it's sub-member in list index order
                value = (4, tuple(value))

        return value

    def sort(self, key_or_list, direction=None):
        """
        Sorts a cursor object based on the input
        :param key_or_list: a list/tuple containing the sort specification,
        i.e. ('user_number', -1), or a basestring
        :param direction: sorting direction, 1 or -1, needed if key_or_list
                          is a basestring
        :return:
        """

        # checking input format

        sort_specifier = list()
        if isinstance(key_or_list, list):
            if direction is not None:
                raise ValueError('direction can not be set separately '
                                 'if sorting by multiple fields.')
            for pair in key_or_list:
                if not isinstance(pair, (list, tuple)):
                    raise TypeError('key pair should be a list or tuple.')
                if not len(pair) == 2:
                    raise ValueError('Need to be (key, direction) pair')
                if not isinstance(pair[0], str):
                    raise TypeError('first item in each key pair must '
                                    'be a string')
                if not isinstance(pair[1], int) or not abs(pair[1]) == 1:
                    raise TypeError('bad sort specification.')

            sort_specifier = key_or_list

        elif isinstance(key_or_list, str):
            if direction is not None:
                if not isinstance(direction, int) or not abs(direction) == 1:
                    raise TypeError('bad sort specification.')
            else:
                # default ASCENDING
                direction = 1

            sort_specifier = [(key_or_list, direction)]

        else:
            raise ValueError('Wrong input, pass a field name and a direction,'
                             ' or pass a list of (key, direction) pairs.')

        # sorting

        _cursordat = self.cursordat

        total = len(_cursordat)
        pre_sect_stack = list()
        for pair in sort_specifier:

            is_reverse = bool(1-pair[1])
            value_stack = list()
            for index, data in enumerate(_cursordat):

                # get field value

                not_found = None
                for key in pair[0].split('.'):
                    not_found = True

                    if isinstance(data, dict) and key in data:
                        data = copy.deepcopy(data[key])
                        not_found = False

                    elif isinstance(data, list):
                        if not is_reverse and len(data) == 1:
                            # MongoDB treat [{data}] as {data}
                            # when finding fields
                            if isinstance(data[0], dict) and key in data[0]:
                                data = copy.deepcopy(data[0][key])
                                not_found = False

                        elif is_reverse:
                            # MongoDB will keep finding field in reverse mode
                            for _d in data:
                                if isinstance(_d, dict) and key in _d:
                                    data = copy.deepcopy(_d[key])
                                    not_found = False
                                    break

                    if not_found:
                        break

                # parsing data for sorting

                if not_found:
                    # treat no match as None
                    data = None

                value = self._order(data, is_reverse)

                # read previous section
                pre_sect = pre_sect_stack[index] if pre_sect_stack else 0
                # inverse if in reverse mode
                # for keeping order as ASCENDING after sort
                pre_sect = (total - pre_sect) if is_reverse else pre_sect
                _ind = (total - index) if is_reverse else index

                value_stack.append((pre_sect, value, _ind))

            # sorting cursor data

            value_stack.sort(reverse=is_reverse)

            ordereddat = list()
            sect_stack = list()
            sect_id = -1
            last_dat = None
            for dat in value_stack:
                # restore if in reverse mode
                _ind = (total - dat[-1]) if is_reverse else dat[-1]
                ordereddat.append(_cursordat[_ind])

                # define section
                # maintain the sorting result in next level sorting
                if not dat[1] == last_dat:
                    sect_id += 1
                sect_stack.append(sect_id)
                last_dat = dat[1]

            # save result for next level sorting
            _cursordat = ordereddat
            pre_sect_stack = sect_stack

        # done

        self.cursordat = _cursordat

        return self

    def hasNext(self):
        """
        Returns True if the cursor has a next position, False if not
        :return:
        """
        cursor_pos = self.cursorpos + 1

        try:
            self.cursordat[cursor_pos]
            return True
        except IndexError:
            return False

    def next(self):
        """
        Returns the next record
        :return:
        """
        self.cursorpos += 1
        return self.cursordat[self.cursorpos]

    @property
    def count(self):
        """
        Returns the number of records in the current cursor
        :return: number of records
        """
        return len(self.cursordat)

    def first(self):
        return self.cursordat[0] if len(self.cursordat) else None

    def last(self):
        return self.cursordat[:-1] if len(self.cursordat) else None


def query(data:list, filters:dict) -> List[dict]:
    return DictQuery(data).execute(filters) if filters and data else data



import pytest
from angola import dict_mutator

def test_mutate_simple_set():
  init_data = {
    "name": "mutate",
    "version": "1.0.2"
  }

  mutations = {
    "version": "2.0.1",
    "location": "CLT"
  }

  d, _ = dict_mutator.mutate(mutations)
  assert d["version"] == "2.0.1"

def test_incr():
  init_data = {
    "counter": 2
  }
  m = {
    "counter:$incr": True
  }

  d, _ = dict_mutator.mutate(mutations=m, init_data=init_data)
  assert d.get("counter") == 3

  m = {
    "counter:$incr": 5
  }
  d, _ = dict_mutator.mutate(mutations=m, init_data=init_data)
  assert d.get("counter") == 7

  m = {
    "counter:$incr": -3
  }
  d, _ = dict_mutator.mutate(mutations=m, init_data=init_data)
  assert d.get("counter") == -1
 

def test_decr():
  init_data = {
    "counter": 2
  }
  m = {
    "counter:$decr": True
  }

  d, _ = dict_mutator.mutate(mutations=m, init_data=init_data)
  assert d.get("counter") == 1

  m = {
    "counter:$decr": 5
  }
  d, _ = dict_mutator.mutate(mutations=m, init_data=init_data)
  assert d.get("counter") == -3

  m = {
    "counter:$decr": -3
  }
  d, _ = dict_mutator.mutate(mutations=m, init_data=init_data)
  assert d.get("counter") == 5


# ------------------------------------------------------------------------------
# -- Angola --
# ------------------------------------------------------------------------------

from .database import Database as db, Collection, CollectionItem, CollectionActiveRecordMixin
from .lib_xql import xql_to_aql
from .dict_mutator import mutate as parse_dict_mutations
from .lib import gen_xid
from arango.exceptions import ArangoError



#-----------------------------
# -- Angola --
#-----------------------------

import copy 
from typing import Any, List
from arango import ArangoClient
from arango.exceptions import ArangoError, DocumentUpdateError
from contextlib import contextmanager
from . import lib, lib_xql, dict_mutator, dict_query


DEFAULT_INDEXES = [
    {
        "type": "persistent",
        "fields": ["_created_at"],
        "name": "idx00__created_at"
    },
    {
        "type": "persistent",
        "fields": ["_modified_at"],
        "name": "idx00__modified_at"
    },
    {
        "type": "ttl",
        "fields": ["__ttl"],
        "name": "idx00__ttl",
        "expireAfter": 0
    }
]

#------------------------------------------------------------------------------
# Exception
class AngolaError(Exception): pass
class AdapterError(AngolaError): pass
class CollectionNotFoundError(AngolaError): pass
class CollectionExistsError(AngolaError): pass
class ItemNotFoundError(AngolaError):pass
class ItemExistsError(AngolaError):pass
class NoResultsError(AngolaError): pass
class ConstraintError(AngolaError): pass
class UndeletableError(AngolaError): pass
class MissingCommitterCallbackError(AngolaError): pass
class MissingItemKeyError(AngolaError): pass
class InvalidItemPathError(AngolaError): pass 

#------------------------------------------------------------------------------

class CollectionActiveRecordMixin(object):
    """
    CollectionActiveRecordMixin

    An abstraction class to use as an active record on the items

    Usage:

      class User(CollectionActiveRecordMixin):
        def full_name(self):
            return "%s %s" % (self.get("first_name), self.get("last_name"))
            
      coll = #db.select_collection(..., item_class=User)

      if item := coll.get(_key):
        print(item.full_name())

    """
    _item = None

    def __init__(self, _item, *a, **kw):
        self._item = _item


    def __call__(self, _item=None, *a, **kw):
        i = copy.deepcopy(self)
        i._item = _item
        return i

    def __getattr__(self, __name: str, *a, **kw):
      return self._item.__getattribute__(__name, *a, **kw)

   
#------------------------------------------------------------------------------

class _QueryResultIterator(object):
    results:list = []
    pagination:dict = {}
    cursor:list = []
    count:int = 0
    size:int = 0

    def __init__(self, cursor, pager, data_mapper=None):
        self.cursor = cursor
        stats = cursor.statistics()
        self.count = self.cursor.count() # current count
        self.size = stats["fullCount"] # total count
        self.pagination = lib.gen_pagination(
                                            size=self.size,
                                            count=self.count,
                                            page=pager[0],
                                            per_page=pager[1])

        def _default_data_mapper_cb(item): return item 
    
        _data_mapper = _default_data_mapper_cb if not data_mapper else data_mapper

        self.results = [_data_mapper(item) for item in self.cursor]

    def __iter__(self):
        """
        Iterate over the results
        """
        yield from self.results

    def __len__(self):
        """
        Get the total results 
        """
        return self.size

class _ItemMixin(dict):
    NAMESPACE = None
 
    def _make_path(self, path):
        # if self.NAMESPACE:
        #     return "%s.%s" % (self.NAMESPACE, path)
        return path

    def _update(self, data):
        raise NotImplementedError()

    def get(self, path: str, default: Any = None) -> Any:
        """
        GET: Return a property by key/DotNotation

        ie: 
            #get("key.deep1.deep2.deep3")

        Params:
            path:str - the dotnotation path
            default:Any - default value 

        Returns:
            Any
        """
        path = self._make_path(path)
        return lib.dict_get(obj=dict(self), path=path, default=default)

    def set(self, path: str, value: Any):
        """
        SET: Set a property by key/DotNotation

        Params:
            path:str - the dotnotation path
            value:Any - The value

        Returns:
            Void
        """

        path = self._make_path(path)
        self._update({path: value})
        return self

    def len(self, path: str):
        """
        Get the length of the items in a str/list/dict
        Params:
            path:str - the dotnotation path
        Returns:
            data that was removed
        """
        path = self._make_path(path)
        v = self.get(path)
        return len(v) if v else 0

    def incr(self, path: str, incr=1):
        """
        INCR: increment a value by 1
        Args
            path:str - path
            incr:1 - value to inc by
        Returns:    
            int - the value that was incremented
        """
        op = "%s:$incr" % self._make_path(path)        
        oplog = self._update({op: incr})
        return oplog.get(op)

    def decr(self, path: str, decr=1):
        """
        DECR: decrement a value by 1
        Args
            path:str - path
            decr:1 - value to dec by
        Returns:    
            int - the value that was decremented
        """
        op = "%s:$decr" % self._make_path(path)

        oplog = self._update({op: decr})
        return oplog.get(op)

    def unset(self, path: str):
        """ 
        UNSET: Remove a property by key/DotNotation and return the value

        Params:
            path:str

        Returns:
            Any: the value that was removed
        """
        path = self._make_path(path)
        self._update({"%s:$unset" % path: True})

    def rename(self, path: str, value:str):
        """ 
        RENAME: Rename a property by key/DotNotation and return the value

        Params:
            path:str - The source
            value:str - the target value

        Returns:
            Any: the value that was removed
        """
        path = self._make_path(path)
        self._update({"%s:$rename" % path: value})

    def copy(self, path: str, value:str):
        """ 
        COPY: Copy a property by key/DotNotation and return the value

        Params:
            path:str - The source
            value:str - the target value
        """
        path = self._make_path(path)
        self._update({"%s:$copy" % path: value})

    def xadd(self, path: str, values):
        """
        XADD: Add *values if they don't exist yet

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xadd" % self._make_path(path)
        self._update({op: values})

    def xadd_many(self, path: str, *values: List[Any]):
        """
        XADD: Add *values if they don't exist yet

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xadd_many" % self._make_path(path)
        self._update({op: values})

    def xrem(self, path: str, values):
        """
        XREM: Remove items from a list

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xrem" % self._make_path(path)
        oplog = self._update({op: values})
        return oplog.get(op)

    def xrem_many(self, path: str, *values: List[Any]):
        """
        XREM: Remove items from a list

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xrem_many" % self._make_path(path)
        oplog = self._update({op: values})
        return oplog.get(op)

    def xpush(self, path: str, values: Any):
        """
        XPUSH: push item to the right of list. 

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xpush" % self._make_path(path)
        self._update({op: values})

    def xpush_many(self, path: str, *values: List[Any]):
        """
        XPUSH_MANY: push item to the right of list. 

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xpush_many" % self._make_path(path)
        self._update({op: values})

    def xpushl(self, path: str, values: Any):
        """
        XPUSHL: push item to the right of list. 

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xpushl" % self._make_path(path)
        self._update({op: values})

    def xpushl_many(self, path: str, *values: List[Any]):
        """
        LPUSH: push item to the right of list. 

        Params:
            path:str - the dotnotation path
            *values: set of items
        Returns:
            list: updated data
        """
        op = "%s:$xpush_many" % self._make_path(path)
        self._update({op: values})

    def xpop(self, path: str):
        """
        Remove value at the end an array/list
        Params:
            path:str - the dotnotation path
        Returns:
            data that was removed

        """
        op = "%s:$xpop" % self._make_path(path)
        oplog = self._update({op: True})
        return oplog.get(op)

    def xpopl(self, path: str):
        """
        Remove value at the beginning an array/list
        Params:
            path:str - the dotnotation path
        Returns:
            data that was removed        
        """
        op = "%s:$xpopl" % self._make_path(path)
        oplog = self._update({op: True})
        return oplog.get(op)

    def timestamp(self, path:str, value:Any=True):
        op = "%s:$datetime" % self._make_path(path)
        oplog = self._update({op: value})
        return oplog.get(op)   

    def template(self, path:str, value:str):
        op = "%s:$template" % self._make_path(path)
        oplog = self._update({op: value})
        return oplog.get(op)

    def uuid4(self, path:str):
        op = "%s:$uuid4" % self._make_path(path)
        oplog = self._update({op: True})
        return oplog.get(op)

    def update(self, data: dict) -> "Self":
        """
        UPDATE: Update the active CollectionItem

        Returns:
            CollectionItem

        Example:
            #item.update({k/v...})
            #item.commit()

            or 
            #item.update({k/v...}).commit()
        """
        self._update(data)
        return self

class CollectionItem(_ItemMixin):
    """
    CollectionItem

    Every row is a document 
    """

    # item _key
    _key = None

    # items subcollections
    _subcollections = {}
    
    # immutable keys
    _immut_keys = []

    # _read_only. When read_only, it can update 
    _read_only = False

    @classmethod
    def new(cls, data:dict, immut_keys:list=[], db=None, collection=None, commiter=None, custom_ops:dict={}, read_only:bool=False):
      return cls(data=_create_document_item(data), db=db, collection=collection, immut_keys=immut_keys, commiter=commiter, custom_ops=custom_ops, read_only=read_only)

    def __init__(self, data: dict, db=None, collection=None, immut_keys:list=[], load_parser=None, commiter=None, custom_ops:dict={}, read_only:bool=False):
        if "_key" not in data:
            raise MissingItemKeyError()
        self._db = db
        self._collection = collection
        self._load_parser = load_parser
        self._commiter = commiter
        self._immut_keys = immut_keys
        self._cx = False
        self._custom_ops = custom_ops
        self._read_only = read_only
    
        data, _ = dict_mutator.mutate(mutations=data,  immuts=immut_keys, custom_ops=self._custom_ops)
        self._load(data)

    def to_dict(self):
        data = dict(self)
        if self._subcollections:
            data["__subcollections"] = self._subcollections
        return data 

    def set_immut_keys(self, immut_keys:list=[]):
        self._immut_keys = immut_keys

    @contextmanager
    def context(self):
        """
        *ContextManager for CollectionItem

        Do transactional mutation and commit the changes upon exit

        Yield:
            CollectionItem

        Example:
            with item.context() as ctx:
                ctx.update({"name": "Y"})

        """
        yield self 
        self.commit()

    @contextmanager
    def context_subcollection(self, name: str, constraints: list = None):
        """
        *Context Manager for Subcollection

        Do transactional mutation on subcollection and commit the changes upon exit

        Yield:
          SubCollection

        Example:

        with item.context_subcollection('name') as sc:
            sc.insert()
        
        """
        sc = SubCollection(item=self, name=name, custom_ops=self._custom_ops, constraints=constraints)
        yield sc
        self.commit()

    def select_subcollection(self, name: str, constraints: list = None):
        """

        Select a subcollection. When making changes, must use `commit` on parent

        Retuns:
          SubCollection

        Example:
            sc = item.select_subcollection(name)
            sc.insert({...})
            sc.insert({...})
            item.commit()

            -- or with #context
            with item.context() as ictx:
                sc = item.select_subcollection(name)
                sc.insert({...})
                sc.insert({...})
            
            or refer to #context_subscollection

        """
        return SubCollection(item=self, name=name, custom_ops=self._custom_ops, constraints=constraints)

    def get_item(self, path:str) -> "_SubCollectionItem":
        """
        To get a subcollection item via path

        Path: [SUB_COLLECTION_NAME/DOCUMENT_KEY] -> articles/1234568

        Params:
            path:str - str of [sub_collection_name/document_key]
        Return:
            collection.item

        Example:
            db.get_item("collection/_key").get_item("sub_collection/_key")
        
        Returns:
            _SubCollectionItem

        """
        
        paths = path.split("/")
        if len(paths) != 2:
            raise InvalidItemPathError()

        return self.select_subcollection(paths[0]).get(paths[1])        

    @property
    def subcollections(self) -> list:
        """ List all collections """
        return list(self._subcollections.keys())

    def drop_subcollection(self, name: str):
        try:
            if name in self._subcollections:
                del self._subcollections[name]
            self.set("__subcollections", self._subcollections)
        except KeyError as _:
            pass
        return True

    def _set_subcollection(self, name:str, data:Any):
        self._subcollections[name] = data
        self.set("__subcollections", self._subcollections)

    def commit(self) -> "Self":
        """ To save """

        if self._read_only:
            return self
        
        if not self._commiter:
            raise MissingCommitterCallbackError()
        data = self._commiter(self)
        if data:
            self._load(data)
        return self
        
    def _update(self, mutations: dict):
        """
        Return oplog
        """
        
        if self._read_only:
            return self
        
        data = self.to_dict()
        doc, oplog = dict_mutator.mutate(mutations=mutations, init_data=data, immuts=self._immut_keys, custom_ops=self._custom_ops)
        self._load(doc)
        return oplog

    def _load(self, item: dict):
        """
        load the content into the document

        Params:
            row: dict
        """
        self._clear_self()
        
        if self._load_parser:
          item = self._load_parser(item)

        self._subcollections = {}
        if "__subcollections" in item:
            self._subcollections = item.pop("__subcollections") or {}

        if "_key" in item:
            self._key = item.get("_key")
        super().__init__(item)

    def _clear_self(self):
        """ clearout all properties """
        for _ in list(self.keys()):
            if _ in self:
                del self[_]

    def link(self, to_item:"CollectionItem", data={}, edge_collection_name=None):
        """
        ::Graph::

        Create a graph link between this item and to_item
        """
        return self._db.link_edges(from_item=self, to_item=to_item, data=data, edge_collection_name=edge_collection_name)

    def traverse(self, collection:"Collection", relations:list("Collection")=None, direction="outbound"): 
        return self._db.traverse(from_item=self, collection=collection, relations=relations, direction=direction)

    def set_ttl(self, nattime:str) -> "Self":
        """
        To set a time to live on an item
        ie: 
            item.set_ttl("2days")

            to reset the ttl
                item.set_ttl(False)

        Params:
            - nattime:str - Natural time - ie 1hour, 60seconds

        Returns:
            self
        """
        if isinstance(nattime, str):
            self.update({"__ttl:$datetime": nattime})
        elif nattime is False:
            self.update({"__ttl": None})
        return self

    def delete(self) -> bool:
        """
        To delete an item
        Todo: deleted links/edges
        """
        self._collection.delete(self._key)
        return True

    def get_sizeof(self) -> int:
        """
        Get the size of the document

        Returns: int
        """
        return lib.get_sizeof(self.to_dict())

class _SubCollectionItem(_ItemMixin):
    _key = None 

    def __init__(self, subCollection: "SubCollection", data):
        self._subcollection = subCollection
        self._load(data)

    @property
    def parent(self):
        """
        Holds parent data
        """
        return self._subcollection._item

    def _update(self, mutations):
        data = dict(self)
        mutations = copy.deepcopy(mutations)
        doc, oplog = dict_mutator.mutate(mutations=mutations, init_data=data, immuts=self.parent._immut_keys, custom_ops=self._subcollection._custom_ops)
        self._subcollection._save(self._key, doc)
        self._load(doc)
        return oplog

    def _load(self, data):
        self._key = data.get("_key")
        super().__init__(data)

    def delete(self):
        self._subcollection.delete({"_key": self._key})
        return True
     
#------------------------------------------------------------------------------
#------------------------------------------------------------------------------

class Database(object):
    """
    Angola Database
    """
    SYSTEM_DB = "_system"

    def __init__(self,
                 hosts:str=None,
                 username:str="root",
                 password:str=None, 
                 dbname: str = SYSTEM_DB, 
                 client:"Database"= None, 
                 default_indexes:list=[],
                 query_max_limit=100,
                 collection_prefix:str=None,
                 custom_ops:dict={}):
        """
        
        Params:
            host:str|list
            username:str
            password
            dbname
            client:Database
            default_indexes:list
            query_max_limit
            collection_prefix:str|function - a prefix to add in all collection name
            custom_ops:dict - 
        
        """

        self.client = client
        self.username = username
        self.password = password
        self.db = None
        self.dbname = dbname or self.SYSTEM_DB # fallback to _system_db
        self.default_indexes = default_indexes
        self.query_max_limit = query_max_limit
        self._custom_ops = custom_ops
        self._collection_prefix = collection_prefix

        if not self.client:
            self.client = ArangoClient(hosts=hosts, serializer=lib.json_ext.dumps, deserializer=lib.json_ext.loads)

        if self.dbname:
            self.db = self.client.db(name=self.dbname, username=self.username, password=self.password)

    @property
    def aql(self):
        return self.db.aql

    def prefix_collection_name(self, collection_name:str) -> str:
        if self._collection_prefix:
            if isinstance(self._collection_prefix, str):
                collection_name = "%s%s" % (self._collection_prefix, collection_name)
            elif callable(self._collection_prefix):
                collection_name = self._collection_prefix(collection_name)
        return collection_name

    def unprefix_collection_name(self, collection_name:str) -> str:
        if self._collection_prefix:
            return collection_name.replace(self._collection_prefix, "")
        return collection_name
        
    def has_db(self, dbname:str=None) -> bool:
        """
        Check if the system has a database

        Params:
            dbname:str|None - The dbname to check or the current self.dbname

        Returns: 
            bool
        """
        _dbname = dbname or self.dbname
        sys_db = self.select_db("_system")
        return sys_db.db.has_database(_dbname)

    def create_db(self, dbname:str=None) -> "Database":
        """
        Create a database if doesn't exists
        Params:
            dbname:str|None - The dbname to check or the current self.dbname

        Returns:
            Database
        """
        _dbname = dbname or self.dbname
        sys_db = self.select_db("_system")
        if not sys_db.db.has_database(_dbname):
            sys_db.db.create_database(_dbname)
        return self.select_db(_dbname)

    def select_db(self, dbname:str, collection_prefix:str=None, default_indexes:dict=None) -> "Database":
        """
        Select a different DB using the same connection

        Params:
            dbname:str - The dbname to check
            collection_prefix
            default_indexes
        Returns: 
            Database
        """
        return Database(client=self.client, 
                        dbname=dbname, 
                        username=self.username, 
                        password=self.password, 
                        collection_prefix=collection_prefix or self._collection_prefix, 
                        custom_ops=self._custom_ops, 
                        default_indexes=default_indexes or self.default_indexes,
                        query_max_limit=self.query_max_limit)

    def has_collection(self, collection_name, use_prefix=True) -> bool:
        """
        Test if collection exists in the current db. 

        Params:
            collection_name:str - the collection name 

        Returns:
            bool
        """
        if use_prefix:
            collection_name = self.prefix_collection_name(collection_name)
        return self.db.has_collection(collection_name)

    def create_collection(self, collection_name:str, indexes:list=[], use_prefix=True) -> bool:
        """
        Create a collection if not exists
        Returns: bool
        """
        if not self.has_collection(collection_name):
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            col = self.db.create_collection(collection_name)

            # indexes
            if _indexes := {l.get("name"):l for l in [*DEFAULT_INDEXES, *(self.default_indexes or []), *(indexes or [])]}.values():
                for index in _indexes:
                    col._add_index(index) 
                    
            return True 
        return False

    def select_collection(self, collection_name:str, indexes:list=[], immut_keys:list=[], item_class=None, auto_create:bool=True, use_prefix=True) -> "Collection":
        """
        To select a collection

        Params:
            collection_name:str - collectioin name 
            indexes:List[dict] - the indexes to use
            immut_keys:list - immutable keys. Keys that can't be updated once created
            auto_create:bool - To auto create the collection if doesn't exist
            use_prefix:bool - To use the prefix or not
        Return: Collection

        """

        if self.has_collection(collection_name):
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            col = self.db.collection(collection_name)
        elif auto_create is True:
            self.create_collection(collection_name=collection_name, indexes=indexes)
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            col = self.db.collection(collection_name)
        else:
            raise CollectionNotFoundError()

        return Collection(db=self, collection=col, immut_keys=immut_keys, custom_ops=self._custom_ops, item_class=item_class)

    def select_edge_collection(self, collection_name:str, use_prefix=True):
        if self.db.has_collection(collection_name):
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            return self.db.collection(collection_name)
        else:
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            return self.db.create_collection(name=collection_name, edge=True)

    def get_item(self, path:str) -> CollectionItem:
        """
        To get an item via path.

        - Item Path: [COLLECTION_NAME/KEY] -> articles/1234568
        - Item's Subcollection path: [COLLECTION/_KEY/SUBCOLLECTION] -> articles/1234/comments
        - Item's Subcollection sub Item path: [COLLECTION/_KEY/SUBCOLLECTION/_SUB_KEY] -> articles/1234/comments/73992

        Params:
            path:str - str of [collection_name/document_key]
        Return:
            collection.item

        Example:
            db.get_item("articles/somethingf")
        
        Returns:
            CollectionItem

        """

        paths = path.split("/")
        len_paths = len(paths)
        if len_paths < 2 or len_paths > 4:
            raise InvalidItemPathError()

        try:
            if len_paths == 2: # item -> [coll/key]
                return self.select_collection(self.prefix_collection_name(paths[0])).get(paths[1])
            elif len_paths == 3: # item's subcolelction -> [coll/key/subcoll]
                return self.select_collection(self.prefix_collection_name(paths[0])).get(paths[1]).select_subcollection(paths[2])
            elif len_paths == 4: # item's subcollection items -> [coll/key/subcoll/subkey]
                return self.select_collection(self.prefix_collection_name(paths[0]))\
                    .get(paths[1])\
                    .select_subcollection(paths[2])\
                    .get(paths[3])
        except Exception as e:
            return None

    def execute_aql(self, query:str, bind_vars:dict={}, *a, **kw):
        """ 
        Execute AQL 
        Params:
            query:str - the AQL to execute 
            bind_vars: dict - the variables to pass in the query
        Return aql cursor
        """
        return self.aql.execute(query=query, bind_vars=bind_vars, *a, **kw)

    def query(self, xql:lib_xql.XQLDEFINITION, data:dict={}, parser=None, data_mapper=None) -> _QueryResultIterator:
        """
        XQL query  a collection based on filters

        It will return the cursor:ArangoCursor and a pagination for the current state
        
        Params:
            xql:lib_xql.XQLDEFINITION
            data:dict
            data_mapper:function - a callback function
        Returns
            _QueryResultIterator
        """

        aql, bind_vars, pager = self.build_query(xql=xql, data=data, parser=parser)
        cursor = self.execute_aql(aql, bind_vars=bind_vars, count=True, full_count=True)            
        return _QueryResultIterator(cursor=cursor, pager=pager, data_mapper=data_mapper)

    def build_query(self, xql:lib_xql.XQLDEFINITION, data:dict={}, parser=None):
        """
        Build a query from XQL

        Return tuple:
            - aql:str
            - bind_vars:dict
            - pagination:tuple 
                -> tuple(page, per_page)
        """        
        xql = lib_xql.prepare_xql(xql)

        # pagination
        if "page" in data:
            xql["PAGE"] = data.get("page") or 1
            del data["page"]
        if "limit" in data:
            xql["LIMIT"] = data.get("limit") or 10
            del data["limit"]

        _per_page, _, _page = lib_xql.xql_take_skip_page(xql=xql, max_limit=self.query_max_limit)
        aql, bind_vars = lib_xql.xql_to_aql(xql, vars=data, parser=parser, max_limit=self.query_max_limit)
        bind_vars.update(data)
        return aql, bind_vars, (_page, _per_page)

    def return_count(self, xql:lib_xql.XQLDEFINITION, data:dict={}) -> int:
        """
        From a query, it returns the count 

        Params:
            xql:lib_xql.XQLDEFINITION
            data:dict
        Returns
            Interger
        """
        xql["RETURN_COUNT"] = True
        aql, bind_vars, pager = self.build_query(xql=xql, data=data)
        cursor = self.execute_aql(aql, bind_vars=bind_vars, count=True, full_count=True) 
        return cursor.batch()[0] if cursor.count() else 0

    def collections(self) -> list:
        """
        All collections in the db

        Returns:
            list
        """
        return self.db.collections()
    
    def rename_collection(self, collection_name:str, new_name:str, use_prefix=True):
        """
        Rename collection
        """
        if self.has(collection_name, use_prefix=use_prefix) and not self.has(new_name, use_prefix=use_prefix):
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            if use_prefix:
                new_name = self.prefix_collection_name(new_name)
            coll = self.select_collection(collection_name, use_prefix=use_prefix)
            coll.collection.rename(new_name)
            return self.select_collection(new_name, use_prefix=use_prefix)
    
    def delete_collection(self, collection_name:str, use_prefix=True):
        if self.has_collection(collection_name, use_prefix=use_prefix):
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            self.db.delete_collection(collection_name)

    def truncate_collection(self, collection_name:str, use_prefix=True):
        if self.has_collection(collection_name, use_prefix=use_prefix):
            if use_prefix:
                collection_name = self.prefix_collection_name(collection_name)
            coll = self.select_collection(collection_name, use_prefix=use_prefix)
            coll.collection.truncate()


    def add_index(self, collection_name, data:dict):
        """
        Args:
            - collection, the collection name
            - data: dict of 
                    {
                        "type": "persistent",
                        "fields": [] # list of fields
                        "unique": False # bool - Whether the index is unique
                        "sparse": False # bool,
                        "name": "" # str - Optional name for the index
                        "inBackground": False # bool - Do not hold the collection lock
                    }
        """
        collection_name = self.prefix_collection_name(collection_name)
        col = self.db.collection(collection_name)
        col._add_index(data)
      
    def delete_index(self, collection_name:str, id):
        """
        Delete Index

        Args:
            - collection, the collection name
            - id: the index id
        """
        collection_name = self.prefix_collection_name(collection_name)
        col = self.db.collection(collection_name)
        col.delete_index(id, ignore_missing=True)

    def link_edges(self, from_item:"CollectionItem", to_item:"CollectionItem", data:dict={}, edge_collection_name:str=None, use_prefix=True):
        """
        GRAPH
        Link edges of 2 collections from:to

        Params:
            - from_item:CollectionItem
            - to_item:CollectionItem
            - data:dict
            - edge_collection_name:str
        """
        from_coll_name = from_item._collection.name
        to_coll_name = to_item._collection.name
        from_cname = self.unprefix_collection_name(from_coll_name)
        to_cname = self.unprefix_collection_name(to_coll_name)

        if not edge_collection_name and from_cname and to_cname:
            edge_collection_name = _create_edge_name(from_cname, to_cname)

        if not edge_collection_name:
            raise Exception("MISSING EDGE COLLECTION NAME")

        coll = self.select_edge_collection(edge_collection_name, use_prefix=use_prefix)

        # ensuring _id and _key are not included
        if data:
            for k in ["_id", "_key"]:
                if k in data:
                    data.pop(k)
        doc = {
            **data,
            "_from": "%s/%s" % (from_coll_name, from_item._key),
            "_to": "%s/%s" % (to_coll_name, to_item._key)
        }

        # update edge if exists, otherwise insert
        edge = list(coll.find({"_from": doc["_from"], "_to": doc["_to"]}, limit=1))
        if edge:
            return coll.update({**data, "_id": edge[0]["_id"]})
        else:
            return coll.insert(doc)

    def unlink_edges(self, from_item:"CollectionItem", to_item:"CollectionItem"):
        pass

    def _load_item(self, data:dict) -> "CollectionItem":
        if "_id" in data:
            collection_name, _key = data.get("_id").split("/")
            collection_name = self.prefix_collection_name(collection_name)
            col = self.db.collection(collection_name)
            return Collection(db=self, collection=col, custom_ops=self._custom_ops).item(data)
        else:
            raise Exception("INVALID_ITEM__MUST_HAVE_ID")

    def traverse(self, from_item:"CollectionItem", collection:"Collection", relations:list("Collection")=None, direction="outbound", filters={}):
        """
        == GRAPH == 
        To traverse 

        Params:
            - from_item:CollectionItem - The item to traverse from
            - collection: Collection
            - relations:list[Collection.link]

        Example:
            item = $coll.get("_key")
            traversial = $coll.traverse(from_item=item, collection=$collectionInstance2)
            for e in traversial:
                item_, second_item = e

            #== Advanced, 4 depth. # the deeper relation the more the data
            item = $coll.get("_key")
            relations = [ $collInstance2.link($collInstance3), $collInstance3.link($collInstance4)]
            traversial = $coll.traverse(from_item=item, collection=$collectionInstance2)
            for e in traversial:
                item_, second_item, third_item, fourth_item = e
            


        AQL:
            FOR v, e, p in 2..3 OUTBOUND "test_country/01gm1g1wed20nf0d835m8k963k" 
            GRAPH "graph__edges__test_country--test_region--test_region--test_city"
            FILTER p.vertices[1].name == "SC"
            RETURN p
        """

        _key = from_item._key

        if not collection:
            raise Exception("Missing")

        min_depth = 1
        defs = []
        edge_name = "edges__%s" % from_item._collection.name
        edge_name += "--%s" % collection.collection_name
        defs.append({
            "edge_collection": _create_edge_name(from_name=from_item._collection.name, to_name=collection.collection_name),
            "from_vertex_collections": [from_item._collection.name],
            "to_vertex_collections": [collection.collection_name]
        })    

        # Add more relations
        if relations:
            min_depth += len(relations)
            for collection_def in relations:
                edge_name += "--%s" % collection_def[0]
                defs.append({
                    "edge_collection": collection_def[1],
                    "from_vertex_collections": [collection_def[2]],
                    "to_vertex_collections": [collection_def[3]]
                })

        graph_name = "graph__%s" % edge_name
        start_vertex = "%s/%s" % (from_item._collection.name, _key)

        if self.db.has_graph(graph_name):
            graph = self.db.graph(graph_name)
        else:
            graph = self.db.create_graph(name=graph_name, edge_definitions=defs)

        for trav in graph.traverse(start_vertex=start_vertex,
                                    direction=direction,
                                    strategy='bfs',
                                    edge_uniqueness='global',
                                    vertex_uniqueness='global',
                                    min_depth=min_depth, 
                                    )["paths"]:
                
                yield tuple(self._load_item(item) for item in trav["vertices"])

    def get_collection_stats(self, collection_name:str, use_prefix=True) -> dict:
        coll = self.select_collection(collection_name=collection_name, use_prefix=use_prefix)
        koll = coll.collection
        count = koll.count()
        statistics = koll.statistics() or {"documents_size": 0, "indexes": {"count": 0, "size": 0}}
        return {
            "documents_count": count, 
            "documents_size": statistics.get("documents_size"),
            "indexes_count": statistics.get("indexes").get("count"),
            "indexes_size": statistics.get("indexes").get("size"),
        }

    
class Collection(object):

    def __init__(self, db:Database, collection,  immut_keys:list=[], custom_ops:dict={}, item_class=None):
        self.db = db
        self.collection = collection
        self._immut_keys = immut_keys
        self._custom_ops = custom_ops
        self.collection_name = self.collection.name
        self.item_class = item_class

    def _commit(self, item:CollectionItem, replace_document=False):
        """
        To commit/save changes

        Params:
            item:CollectionItem
            replace_document:bool - To replace instead of updating. Replace will not merge the data.
        """
        if not item._key:
            raise MissingItemKeyError()
        try:
            if replace_document:
                return self.collection.replace(item.to_dict(), return_new=True)["new"]
            else:
                item.timestamp('_modified_at')
                return self.collection.update(item.to_dict(), return_new=True)["new"]
        except DocumentUpdateError as due:
            item.update({"_modified_at": None})
            return self.collection.insert(item.to_dict(), return_new=True)["new"]

    def __iter__(self):
        return self.find(filters={})

    def item(self, data:dict, read_only:bool=False) -> CollectionItem:
        """
        Load data as item

        Returns:
            CollectionItem
        """
        item = None
        if not isinstance(data, CollectionItem) and "_key" not in data:
            item = CollectionItem.new(data, db=self.db, collection=self.collection, commiter=self._commit, immut_keys=self._immut_keys, custom_ops=self._custom_ops)               
        else:
            item = CollectionItem(data, db=self.db, collection=self.collection, commiter=self._commit, immut_keys=self._immut_keys, custom_ops=self._custom_ops, read_only=read_only)

        return self.item_class(item) if item and self.item_class else item

    def has(self, _key) -> bool: 
        """
        Check if a collection has _key

        Args:
            _key:str

        Returns: 
            Bool
        """
        return self.collection.has(_key)

    def get(self, _key) -> CollectionItem:
        """ 
        Get a document from the collection and returns a collectionItem
        Returns:
            CollectionItem
        """

        if data := self.collection.get(_key):
            return self.item(data)
        return None

    def create(self, data:dict={}) -> CollectionItem:
        """
        To create a new Item without inserting in the collection

        Requires #commit() to save data

        Returns:
            CollectionItem

        Example:
            item = coll.create({...})
            item.commit()
        """
        return CollectionItem.new(data, db=self.db, collection=self.collection, commiter=self._commit, custom_ops=self._custom_ops)

    def insert(self, data:dict, _key=None) -> CollectionItem:
        """
        To insert a new Item and commit in the collection

        Returns:
            CollectionItem

        Example
            item = coll.insert(...)
        """
        if _key or "_key" in data:
            _key = _key or data["_key"]
            if self.has(_key):
                raise ItemExistsError()
            data["_key"] = _key
        item = data
        if not isinstance(data, CollectionItem):
            item = CollectionItem.new(data, db=self.db, collection=self.collection, custom_ops=self._custom_ops)
        self.collection.insert(item.to_dict(), silent=True)
        return self.get(item._key) 

    def update(self, _key:str, data:dict, replace_document=False) -> CollectionItem:
        """
        To update and item. Can also replace the item

        Params:
            _key:str - document key 
            data:dict - data to update
            replace_document:bool - If true will replace the document and not merge

        Returns
            CollectionItem

        """
        item = self.item({**data, "_key": _key})
        self._commit(item, replace_document=replace_document)  
        return self.get(item._key) 

    def upsert(self, data:dict) -> CollectionItem:
        """
        To update or insert data.

        Args:
            data:dict

        Return:
            CollectionItem
        """

        if "_key" in data:
            if self.has(data["_key"]):
                _key = data.pop("_key")
                return self.update(_key=_key, data=data)
        return self.insert(data)

    def delete(self, _key):
        """
        Delete a document by _key
        """
        self.collection.delete(_key)

    def find(self, filters:dict={}, offset=None, limit=10, sort=None, page=None, xql:dict=None):
        """
        Perform a find in the collections

        Returns
            Generator[CollectionItem]
        """
        
        if page is None and offset:
            page = lib.calc_pagination_page_from_offset(offset=offset, per_page=limit)

        elif offset is None and page:
            offset = lib.calc_pagination_offset(page=page, per_page=limit)

        read_only = False
        _xql = {
            "FROM": self.collection_name, 
            "FILTERS": filters,
            "OFFSET": offset,
            "LIMIT": limit,
            "SORT": sort,
            "PAGE": page
        }

        # Extended XQL
        if xql:
            _xql.update(xql)
            if "JOIN" in _xql:
                read_only = True


        def data_mapper(item): return self.item(item, read_only=read_only)
        return self.db.query(_xql, data_mapper=data_mapper)

    def return_count(self, filters:dict={}) -> int:
        """
        Return the total count of documents from query

        Returns
            Int
        """
        
        _xql = {
            "FROM": self.collection_name, 
            "FILTERS": filters
        }
        return self.db.return_count(xql=_xql)

    def find_one(self, filters:dict, sort=None):
        """
        Retrieve one item based on the criteria

        Returns
            CollectionItem
        """
        if data := list(self.find(filters=filters, limit=1, sort=None)):
            return data[0]
        return None

    def edges_with(self, collection:"Collection", use_prefix=True) -> tuple:
        """
        ::GRAPH::

        Create an Edge Collection Definition, especially when building a traversal

        ie:
        relations = [CollectionA.edges_with(CollectionB)]

        Returns
            tuple(name, edge_name, from_collection_name, to_collection_name)
        """
        from_coll_name = self.collection_name
        to_coll_name = collection.collection_name
        from_cname = self.db.unprefix_collection_name(from_coll_name)
        to_cname = self.db.unprefix_collection_name(to_coll_name)

        edge_name = _create_edge_name(from_cname, to_cname)
        name = "%s--%s" % (from_cname, to_cname)
        return (name, edge_name, self.collection_name, collection.collection_name)


class SubCollection(object):
    _data = []
    _constraints = []
    _item = None
    _name = None 

    def __init__(self, item: CollectionItem, name: str, constraints:list=None, custom_ops:dict={}):
        self._item = item
        self._name = name
        self._constraints = constraints
        self._load()
        self._custom_ops = custom_ops

    def _load(self):
        self._data = self._item._subcollections.get(self._name) or []

    def _commit(self):
        self._item._set_subcollection(self._name, self._data)

    def _save(self, _key, data):
        _data = self._normalize_data()
        _data[_key] = data
        self._data = self._denormalize_data(_data)
        self._commit()        

    def _normalize_data(self) -> dict:
        return { d.get("_key"): d for d in self._data}

    def _denormalize_data(self, data:dict) -> list:
        return list(data.values())

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return iter(self.find())

    @property
    def items(self):
        """ 
        Returns an iterator of all documents

        Returns:
            Iterator
        """
        return self.find()

    def has(self, _key):
        return bool(self.find_one({"_key": _key}))

    def insert(self, data: dict, _key:str=None):
        """
        Insert document

        Params:
            data:dict
            _key: to insert with a _key
        """
        data, _ = dict_mutator.mutate(mutations=data.copy(), immuts=self._item._immut_keys, custom_ops=self._custom_ops)

        if self._constraints:
            for c in  self._constraints:
                if c in data:
                    if self.find_one({c: data[c]}):
                        raise ConstraintError("Key: %s" % c)

        if _key or "_key" in data:
            _key = _key or data["_key"]
            if self.has(_key):
                raise ItemExistsError()
            data["_key"] = _key
        item = data

        item = _create_subdocument_item(data)
        self._data.append(item)
        self._commit()
        return _SubCollectionItem(self, item)

    def update(self, filters:dict, mutations: dict, upsert:bool=False):
        """
        Update by filter

        Params:
            filter:dict - filter document criteria
            mutations:dict - changes on the found documents
        """
        _data = self._normalize_data()
        res = self.find(filters)
        if res:
            for item in res:
                ts = lib.get_timestamp()
                _key = item.get("_key")
                _default = {  # ensuring we do some data can't be overwritten
                    "_key": _key,
                    # "_created_at": ts
                }
                upd, _ = dict_mutator.mutate(mutations=mutations, init_data=item, immuts=self._item._immut_keys, custom_ops=self._custom_ops)
                _data[_key] = {**upd, **_default}
            self._data = self._denormalize_data(_data)
            self._commit()

        elif upsert:
            self.insert(mutations)
  

    def delete(self, filters: dict):
        """
        Delete documents based on filters

        Params:
            filters:dict
        """
        _data = self._normalize_data()
        for item in self.find(filters):
            del _data[item.get("_key")]
        self._data = self._denormalize_data(_data)
        self._commit()

    def get(self, _key:str) -> "_SubCollectionItem":
        """
        Return a document from subcollection by id 

        Returns: _SubCollectionItem
        """
        return self.find_one({"_key": _key})

    def find_one(self, filters:dict={}) -> "_SubCollectionItem":
        """
        Return only one item by criteria

        Return:
            dict
        """
        if res := self.find(filters=filters, limit=1):
            return list(res)[0]
        return None 

    def find(self, filters: dict = {}, sort: dict = {}, limit: int = 10,  offset:int=0, page=None) -> dict_query.Cursor:
        """
        Perform a query

        Params:
            filters:
            sort:
            limit:
            offset:
        """

        if offset is None and page:
            offset = lib.calc_pagination_offset(page=page, per_page=limit)

        sort = _parse_sort_dict(sort, False)
        data = [_SubCollectionItem(self, d) for d in dict_query.query(data=self._data, filters=filters)]
        return dict_query.Cursor(data, sort=sort, limit=limit, offset=offset)

    def filter(self, filters: dict = {}) -> dict_query.Cursor:
        """
        Alias to find() but makes it seems fluenty
        
        Returns:
            dict_query:Cursor
        """
        data = dict_query.query(data=self._data, filters=filters)
        return dict_query.Cursor([_SubCollectionItem(self, d) for d in data])


#------------------------------------------------------------------------------

def _create_edge_name(from_name, to_name):
    return "edge--%s-%s" % (from_name, to_name)

def _create_document_item(data:dict={}) -> dict:
    _key = data["_key"] if "_key" in data else lib.gen_key()

    return {
        "_key": _key,
        "_created_at:$datetime": True,
        "_modified_at": None,
        "__ttl": None,
        **data,
    }

def _create_subdocument_item(data:dict={}) -> dict:
    _key = data["_key"] if "_key" in data else lib.gen_key()

    return {
        "_key": _key,
        "_created_at:$datetime": True,
        "_modified_at": None,
        **data,
    }

def _ascdesc(v, as_str=True):
    if as_str:
        if isinstance(v, int):
            return "DESC" if v == -1 else "ASC"
    else:
        if isinstance(v, str):
            return -1 if v.upper() == "DESC" else 1
    return v


def _parse_sort_dict(sort: dict, as_str=True):
    return [(k, _ascdesc(v, as_str)) for k, v in sort.items()]




import pytest
from angola import kango
from kango.database import Database, Collection, Document

def test_create_db():
    db = kango()
    assert isinstance(db, Database)

def test_select_collection():
    db = kango()
    coll = db.select("collname")
    assert isinstance(coll, Collection)
    assert coll.name == "collname"

def test_select_collection_magic():
    db = kango()
    coll = db.collname
    assert isinstance(coll, Collection)

def test_select_collection_name_magic():
    db = kango()
    coll = db.collname
    assert coll.name == "collname"

def test_collection_size():
    db = kango()
    coll = db.collname

    for i in range(5):
        coll.insert({"name": "x"})
    assert coll.size == 5


#-----------------------------
# -- Angola --
#-----------------------------

import os
import re
import sys
import json
import xid
import math
import uuid
import time
import arrow
import string
import random
import secrets
import hashlib
import datetime
from slugify import slugify
from jinja2 import Template
from functools import reduce
from operator import itemgetter
from typing import Iterator, Any

# Date format
FORMAT_ISO_DATE = "YYYY-MM-DD"
FORMAT_ISO_TIME = "HH:mm:ss"
FORMAT_ISO_DATETIME = "YYYY-MM-DD HH:mm:ss"

DATES_FORMAT = {
    "ISO_DATE": FORMAT_ISO_DATE,
    "ISO_TIME": FORMAT_ISO_TIME,
    "ISO_DATETIME": FORMAT_ISO_DATETIME
}


def get_sizeof(obj:Any) -> int:
    """
    Get the memory size of the object in byte.

    Note: make use of `sys.getsizeof` to get the size in memory. Do not use `len`

    Example: 
        get_sizeof('hello') -> 56
        get_sizeof({'name': 'awesome'}) -> 68
        get_sizeof(12) -> 51
        get_size(json.loads($json_string)) -> $x
        
    Return: int 
    """
    return sys.getsizeof(json_dumps(obj))



def collection_name_valid(name):
    """
    Valid a collection name 
    length: 2, 64
    pattern: lowercase, letters and numbers and underscore
    """
    pattern = re.compile(r"^[a-z][a-z0-9\_]{2,64}$")
    return bool(pattern.match(name)) 


def gen_xid() -> str:
    """
    Return XID. 26 Chars
    """
    return str(xid.Xid())

def gen_key() -> str:
    """
    Return XID. 26 Chars
    """
    return gen_xid()
    
def gen_uuid() -> str:
    """
    Return a UUID4 key. 32 chars
    """
    return ("%s" % uuid.uuid4()).replace("-", "")

def gen_name(length:int=8) -> str:
    """
    Generate a random name
    semi-pronounceable, semi-memorable
    """
    consonants = "bcdfghjklmnpqrstvwxyz"
    vowels = "aeiou"
    return "".join(random.choice((consonants, vowels)[i % 2]) for i in range(length))


def gen_number(length:int=4) -> int:
    """
    Generate random number
    """
    return ''.join(random.choice(string.digits) for _ in range(length))


def gen_random_password(length:int=20) -> str:
    """
    Generate a random password
    """
    return secrets.token_urlsafe(length)

def gen_dbname(str_len=8, num_len=4):
    """
    Generate a random database name of 12 chars
    """
    return gen_name(str_len) + gen_number(num_len)

# ====

def sanitize_custom_key(key:str) -> str:
    """
    To sanitize a custom key

    Returns:
        string
    """
    return slugify(key, separator="")

def sanitize_custom_name(name:str) ->str:
    """ Sanitize a custom name """
    return slugify(name, separator="_", max_length=64)
    

def make_scoped_name(scope:str, name:str) -> str:
    """
    Create a scoped name
    """
    name = sanitize_custom_name(name)
    return "%s--%s" % (scope.lower(), name)


def parse_scoped_name(scoped_name:str) -> tuple:
    """
    Parse scoped name to return the scope and name
    Returns:
        (scope, name)
        
    """
    sn = scoped_name.split("--", 1)
    return tuple(sn) if  len(sn) == 2 else (None, sn[0])


# === DATE + TIME 
def get_datetime() -> arrow.Arrow:
    """
    Generates the current UTC timestamp with Arrow date

    ISO FORMAT
    Date    2022-08-13
    Date and time in UTC : 2022-08-13T22:45:03+00:00

    Returns:
      Arrow UTC Now
    """
    return arrow.utcnow()


_tsmapshifts = [
    ("years", ["y", "year", "years"]),
    ("months", ["m", "month", "months"]),
    ("days", ["d", "day", "days"]),
    ("hours", ["hh", "hr", "hrs", "hour", "hours"]),
    ("minutes", ["mm", "min", "minute", "minutes"]),
    ("seconds", ["ss", "sec", "second", "seconds"]),
    ("weeks", ["wk", "wks", "week", "weeks"])
]
TS_MAP_SHIFTS =  {x:m[0] for m in _tsmapshifts for x in m[1]}
VALID_SHIFTS = ["years", "months", "days", "hours", "minutes", "seconds", "weeks"]

def arrow_date_shifter(dt: arrow.Arrow, stmt: str) -> arrow.Arrow:
    """
    To shift the Arrow date to future or past

    Args:
        dt:arrow.Arrow - 
        stmt:str - 
    Returns:
        arrow.Arrow

    Valid shift:
        YEARS, MONTHS, DAYS, HOURS, MINUTES, SECONDS, WEEKS

    Format: [[+/-][$NUMBER][$SHIFT][SPACE]... ]
        +1Days
        -3Hours 6Minutes
        +1Days 2Hours 3Minutes
        1Year 2Month +3Days 5Hours -6Minutes 3Seconds 5weeks
    """

    t = [t for t in stmt.split(" ") if t.strip(" ")]
    t2 = [re.findall(r'((?:\+|\-)?(?:\d+))?(\w+)?', s)[0] for s in t if s]
    t3 = []
    for l in [(t[1].lower(), int(t[0])) for t in t2 if t[0] and t[1]]:
        if l[0] in VALID_SHIFTS:
            t3.append(l)
        elif l[0] in TS_MAP_SHIFTS:
            t3.append((TS_MAP_SHIFTS.get(l[0]), l[1]))

    if t3:
        dt = dt.shift(**dict(t3))
        return dt

    return dt

def arrow_date_format(dt:arrow.Arrow, format:str='YYYY-MM-DD HH:mm:ss ZZ') -> str:
    if format in DATES_FORMAT:
        format = DATES_FORMAT.get(format)
    return dt.format(format)


def get_timestamp() -> int:
    """
    Generates the current UTC timestamp with datetime
    Returns:
      int
    """
    return round(datetime.datetime.utcnow().timestamp())


# ----------------------
# json_ext

class json_ext:
    """ 
    JSON Extension class to loads and dumps json
    """

    @classmethod
    def dumps(cls, data: dict) -> str:
        """ Serialize dict to a JSON formatted """
        return json.dumps(data, default=cls._serialize)

    @classmethod
    def loads(cls, data: str) -> dict:
        """ Deserialize a JSON string to dict """
        if not data:
            return None
        if isinstance(data, list):
            return [json.loads(v) if v else None for v in data]
        return json.loads(data, object_hook=cls._deserialize)

    @classmethod
    def _serialize(cls, o):
        return cls._timestamp_to_str(o)

    @classmethod
    def _deserialize(cls, json_dict):
        for k, v in json_dict.items():
            if isinstance(v, str) and cls._timestamp_valid(v):
                json_dict[k] = arrow.get(v)
        return json_dict

    @staticmethod
    def _timestamp_valid(dt_str) -> bool:
        try:
            datetime.datetime.fromisoformat(dt_str.replace('Z', '+00:00'))
        except:
            return False
        return True

    @staticmethod
    def _timestamp_to_str(dt) -> str:
        if isinstance(dt, arrow.Arrow):
            return dt.for_json()
        elif isinstance(dt, (datetime.date, datetime.datetime)):
            return dt.isoformat()
        return dt

# alias 
json_dumps = json_ext.dumps
json_loads = json_ext.loads



# === DICT extensions

def flatten_dict(_dict: dict, _str: str = '', reducer=None) -> dict:
    """
    To flatten a dict. Nested node will be separated by dot or separator
    It takes in consideration dict in list and flat them.
    Non dict stay as is

    Args:
        ddict:
        prefix:
    Returns:
        dict

    """
    sep = "."
    ret_dict = {}
    for k, v in _dict.items():
        if isinstance(v, dict):
            ret_dict.update(flatten_dict(v, _str=sep.join([_str, k]).strip(sep)))
        elif isinstance(v, list):
            _k =  ("%s.%s" % (_str, k)).strip(sep)
            ret_dict[_k] = [flatten_dict(item) if isinstance(item, dict) else item for item in v]
        else:
            ret_dict[sep.join([_str, k]).strip(sep)] = v
    return ret_dict


def unflatten_dict(flatten_dict: dict) -> dict:
    """
    To un-flatten a flatten dict

    Args:
      flatten_dict: A flatten dict
    Returns:
      an unflatten dictionnary
    """
    output = {}
    for k, v in flatten_dict.items():
        path = k.split(".")
        if isinstance(v, list):
            v = [unflatten_dict(i2) if isinstance(i2, dict) else i2 for i2 in v]
        _set_nested(output, path, v, k)
    return output


def _get_nested_default(d, path):
    return reduce(lambda d, k: d.setdefault(k, {}), path, d)


def _set_nested(d, path, value, full_path=None):
    try:
        _get_nested_default(d, path[:-1])[path[-1]] = value
    except (AttributeError, TypeError) as e:
        err = "DataTypeAttributeError: %s at '%s'" % (e, full_path)
        raise AttributeError(err)

def dict_pick(ddict: dict, keys: list, check_keys=False) -> dict:
    """
    To pick and return specific keys from a flatten dict.

    Args:
      ddict: dict
      keys: A list of dot notation path to keep
      check_keys: bool - check if all keys exist

    Returns:
      a dict with the picked value

    Example
      keys: ["name", "location.city"]
      flatten_dict: { "name": "MM", "location.city": "Charlotte",
          "location.state": "NC", "age": 100}
      returns: {
        "name": "MM",
        "location": {
          "city": "Charlotte"
        }
      }
    """
    fd = flatten_dict(ddict)

    # check that all keys exist
    if check_keys:
        for k in keys:
            assert k in fd, "missing key '%s'" % k

    ufd = _dict_pick_merge_l2d([_dict_pick_lookup_dict(fd, k) for k in keys])
    return unflatten_dict(ufd)


def _dict_pick_merge_l2d(iter: list):
    """
    Flatten a looked up list of tuples into dict
    Returns dict
    """
    df = {}
    for u in iter:
        if isinstance(u, list):
            for u2 in u:
                df[u2[0]] = u2[1]
    return df


def _dict_pick_lookup_dict(fd, k):
    """
    To lookup a key in a flatten dict 
    If the key is not found directly, it will start from the parent dot

    Returns list of tuples
    """
    if k in fd:
        return [(k, fd[k])]
    p = []
    for dk, dv in fd.items():
        if dk.startswith("%s." % k):
            p.append((dk, dv))
    return p


def dict_find_replace(ddict: dict, kv_repl: dict, is_flatten=False):
    """
    Find/Replace a KV dict in a dict

    Args:
      - ddict
      - kv_repl
      - is_flatten

    Returns
      dict
    """

    fd = ddict
    for k, v in fd.items():
        if isinstance(v, str) and v in kv_repl:
            fd[k] = kv_repl[v]
        elif isinstance(v, list):
            for l, e in enumerate(v):
                if isinstance(fd[k][l], str) and fd[k][l] in kv_repl:
                    fd[k][l] = kv_repl[fd[k][l]]
                elif isinstance(fd[k][l], dict):
                    fd[k][l] = dict_find_replace(fd[k][l], kv_repl)
        elif isinstance(v, dict):
            fd[k] = dict_find_replace(v, kv_repl)
    return fd


def dict_set(my_dict, key_string, value):
    """
    dict_set
    Mutable
    """
    here = my_dict
    keys = key_string.split(".")
    for key in keys[:-1]:
        here = here.setdefault(key, {})
    here[keys[-1]] = value


def dict_get(obj, path, default=None):
    """
    Get a value via dot notaion

    Args:
        @obj: Dict
        @attr: String - dot notation path
            object-path: key.value.path
            object-with-array-index: key.0.path.value
    Returns:
        mixed
    """
    def _getattr(obj, path):
        try:
            if isinstance(obj, list) and path.isdigit():
                return obj[int(path)]
            return obj.get(path, default)
        except:
            return default
    return reduce(_getattr, [obj] + path.split('.'))

def dict_pop(obj:dict, path:str) -> Any:
    """
    * Mutates #obj

    To pop a property from a dict dotnotation

    Args:
        obj:dict - This object will be mutated
        path:str - the dot notation path to update
        value:Any - value to update with

    Returns:
        Any - The value that was removed
    """

    here = obj 
    keys = path.split(".")

    for key in keys[:-1]:
        here = here.setdefault(key, {})
    if isinstance(here, dict):
        return here.pop(keys[-1])
    else:
        val = here[keys[-1]]
        del here[keys[-1]]
        return val



def dict_merge(*dicts) -> dict:                                                            
    """         
    Deeply merge an arbitrary number of dicts                                                                    
    Args:
        *dicts
    Return:
        dict

    Example
        dict_merge(dict1, dict2, dict3, dictN)
    """                                                                             
    updated = {}                                                                    
    # grab all keys                                                                 
    keys = set()                                                                    
    for d in dicts:                                                                 
        keys = keys.union(set(d))                                                   

    for key in keys:                                                                
        values = [d[key] for d in dicts if key in d]                                                                
        maps = [value for value in values if isinstance(value, dict)]            
        if maps:                                                                    
            updated[key] = dict_merge(*maps)                                       
        else:                                                                                                      
            updated[key] = values[-1]                                               
    return updated 

def chunk_list (lst:list, n:int):
    """
    Yield successive n-sized chunks from lst.
    """
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


def dict_upper_keys(obj) -> dict:
    """
    Change all the keys to uppercase in dict
    Args:
        obj: dict
    Returns
        dict

    """
    return { k.upper(): v for k, v in obj }

def list_sorted_dict(l:list, key:str) -> list:
    """
    To order a list of dict, by the value in the dict

    list_sorted_dict(list, "name")
    """
    return sorted(l, key=itemgetter(key))

#===

def hash_string(s):
    return hashlib.sha256(s.encode('utf-8')).hexdigest()


def calc_pagination_offset(page: int, per_page: int) -> int:
    """
    Calculate a pagination offset. 
    To use with LIMIT offset, per_page

    Args:
        - page:int - the current page
        - per_page:int - items per page
    Returns: int 
    """
    return 0 if page < 1 else (page - 1) * per_page

def calc_pagination_page_from_offset(offset:int, per_page:int) -> int:
    """
    To calculate the pagination page having provided an offset and a per_page
    
    Args:
        - offset:int - The offset 
        - per_page: total items per page
    Returns: int
    """
    return 1 if offset <= 0 else round((offset / per_page))


def gen_pagination(size: int, count: int, page: int, per_page: int) -> dict:
    """
    Create pagination data for pagination components

    Args:
        - size:int - The total items
        - count:int - the current count for the subset
        - page:int - the current page
        - per_page:int - total items per page

    Returns: dict
        - page:int - The current page
        - per_page:int - total items per page
        - size:int - total items
        - count:int - the current count for the page
        - page_showing_start:int|None - data start, ie: showing *1 to 10
        - page_showing_end:int|None - data end, ie: showing 1 to *10
        - total_pages:int - total pages
        - has_prev:bool - If it has prev page
        - prev_page:int|None - the prev page # or None
        - has_next:bool - If it has next page
        - next_page:int|None - the next page # or None

    """
    per_page = int(round(per_page))
    if per_page < 1:
        per_page = 10
    total_pages = math.ceil(size / per_page)
    page = int(round(page))
    if page < 1:
        page = 1
    elif page > total_pages:
        page = total_pages
    has_prev = page > 1 and page <= total_pages
    has_next = page < total_pages
    _offset = calc_pagination_offset(page=page, per_page=per_page)
    page_showing_start =  _offset + 1
    page_showing_end = _offset + count
    if total_pages == 0:
        page_showing_start = 0
        page_showing_end = 0
        
    return {
        "page": page,
        "per_page": per_page,
        "count": count,
        "size": size,
        "total_pages": total_pages,
        "has_prev": has_prev,
        "prev_page": page - 1 if has_prev else None,
        "has_next": has_next,
        "next_page": page + 1 if has_next else None,
        "last_page": total_pages,
        "page_showing_start": page_showing_start,
        "page_showing_end": page_showing_end
    }


def render_template(source: str, data={}, is_data_flatten=False) -> str:
    """
    Render Template string with interpolation
    """
    _data = data.copy()
    if _data and is_data_flatten:
        _data = unflatten_dict(_data)
    return Template(source).render(**_data)

def parse_str_template(source: str, data={}):
    """Parse template string with"""
    return Template(source).render(**data)


def parse_dict_template(source: dict, data: dict = {}):
    for k in source:
        if isinstance(source[k], list):
            source[k] = [parse_str_template(i, data) for i in source[k]]
        elif isinstance(source[k], dict):
            source[k] = parse_dict_template(source[k], data)
        else:
            source[k] = parse_str_template(source[k], data)
    return source


def write_template_file(file, data={}):
    """Parse template file with data"""
    with open(file, "r+") as f:
        content = parse_str_template(f.read(), data)
        f.truncate(0)
        f.seek(0)
        f.write(content)



import pytest
from kongodb import lib
import arrow 


def test_gen_id():
    key = lib.gen_id()
    assert len(key) == 26

def test_get_timestamp():
    assert isinstance(lib.get_timestamp(), arrow.Arrow)

def test_dict_get():
    data = {
        "key": "Yo",
        "key2": {
            "key3": "Hello",
            "key4": {
                "location": "NC"
            }
        }
    }
    assert lib.dict_get(data, "key") == "Yo"
    assert lib.dict_get(data, "key2.key3") == "Hello"
    assert lib.dict_get(data, "key2.key4.location") == "NC"


def test_dict_set():
    data = {
        "key": "value"
    }
    lib.dict_set(data, "key", "Loso")
    lib.dict_set(data, "key2.key3", "Hello")
    lib.dict_set(data, "key2.key4", {"location": "NC"})

    assert data["key"] == "Loso"
    assert data["key2"]["key3"] == "Hello"
    assert data["key2"]["key4"]["location"] == "NC"


def test_dict_pop():
    data = {
        "a": { "b": { "c": { "d": 1 } } },
        "aa": { "bb": { "e": { "x": 14 } } }
    }

    p1 = lib.dict_pop(data, "aa.bb.e.x")
    assert p1 == 14
    assert data["aa"] == {"bb": {"e": {}}}
    assert data ==  {'a': {'b': {'c': {'d': 1}}}, 'aa': {'bb': {'e': {}}}}


def test_flatten_dict():
    d1 = {
        "personal": {
            "name": {
                "first": "first name",
                "last": "last name"
            }
        },
        "location": {
            "city": "Charlotte"
        },
        "array": ["A", "B", "C"],
        "list_any": [{"k0": "v0", "d": { "d1" : { "d2": True}}}, 5, "Ok", True]
    }

    f1 = {
        "personal.name.first": "first name",
        "personal.name.last": "last name",
        "location.city": "Charlotte",
        "array": ["A", "B", "C"],
        "list_any": [{'d.d1.d2': True, 'k0': 'v0'}, 5, 'Ok', True]
    }

    assert lib.flatten_dict(d1) == f1


def test_unflatten_dict():
    d1 = {
        "personal": {
            "name": {
                "first": "first name",
                "last": "last name"
            }
        },
        "location": {
            "city": "Charlotte"
        },
        "array": ["A", "B", "C"],
        "list_any": [{"k0": "v0", "d": { "d1" : { "d2": True}}}, 5, "Ok", True]
    }

    f1 = {
        "personal.name.first": "first name",
        "personal.name.last": "last name",
        "location.city": "Charlotte",
        "array": ["A", "B", "C"],
        "list_any": [{'d.d1.d2': True, 'k0': 'v0'}, 5, 'Ok', True]
    }

    assert lib.unflatten_dict(f1) == d1


def test_dict_merge():
    a = {
        'a': 1,
        'b': {
            'b1': 2,
            'b2': 3,
        },
    }
    b = {
        'a': 1,
        'b': {
            'b1': 4,
            'b3': 5
        },
        'c': 6,
    }

    assert lib.dict_merge(a, b)['a'] == 1
    assert lib.dict_merge(a, b)['b']['b2'] == 3
    assert lib.dict_merge(a, b)['b']['b1'] == 4
    assert lib.dict_merge(b, a)['b']['b1'] == 2  # order is flipped
    assert lib.dict_merge(a, b)['b']['b3'] == 5  # new element
    assert lib.dict_merge(a, b)['c'] == 6  # new element


def test_chunk_list():
    data = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
    c = list(lib.chunk_list(data, 3))
    assert isinstance(c, list)
    assert isinstance(c[0], list)
    assert len(c) == 4
    assert len(c[len(c) - 1]) == 2


